{$A+,B-,D-,E-,F+,G-,I-,L+,N-,O+,R-,S-,V-,X-}
UNIT Calc;

{-- Externally available functions --}
INTERFACE
  function Add(a,b : string) : string;
  function Sub(a,b : string) : string;
  function Prod(a,b : string) : string;
  function Divide(    a,b : string;
                  var rm  : string) : string;
  function Fact(var a : string) : string;
  function Power(b,e : string) : string;

{-- Internal functions --}
IMPLEMENTATION
 function SubChar(    c1,c2  : char;
                   var borrow : boolean) : char; assembler;
 {Subtracts one digit char ('0' thru '9') from another and 
 returns the result as a digit. Sets borrow to true if appropriate.}
 ASM
     LES  DI, borrow
     MOV  BYTE PTR ES:[DI], FALSE;
     MOV  AL, c1  
     SUB  AL, c2
     JGE  @NoBorrow
     MOV  BYTE PTR ES:[DI], TRUE
     ADD  AL, 10
 @NoBorrow :
     ADD  AL, 30h
 END;

 function AddChar(    c1, c2 : char;
                   var carry  : boolean) : char; assembler;
 {Adds one digit char ('0' thru '9') to another and returns
 the result as a digit. Sets carry to true if appropriate.}
 ASM
     LES  DI, Carry
     MOV  BYTE PTR ES:[DI], FALSE
     MOV  AL, c1
     ADD  AL, c2
     SUB  AL, 60h  {30h for each digit}
     CMP  AL, 10
     JL   @NoCarry
     SUB  AL, 10
     MOV  BYTE PTR ES:[DI], TRUE
 @NoCarry:
     ADD  AL, 30h
 END;

 function LeftPad0(s   : string;
                     len : byte) : string;
 begin
     if Length(s) < len then
     begin
       Move(s[1], s[Succ(len - Length(s))], Length(s));
       FillChar(s[1], len - Length(s), '0');
     end;
     s[0] := Char(len);
     LeftPad0 := s;
 end;

 procedure TrimLead0(var s : string);
 var p : byte;
 begin
   p := 1;
   while (s[p] = '0') AND (p <= Length(s)) do Inc(p);
   case p of
     0 : s[0] := #0; {string was 255 of '0'! }
     1 : ;           {not found}
   else
     Move(s[p], s[1], Succ(Length(s) - p));
     Dec(s[0], Pred(p));
   end;
 end;

 function Add(a, b : string) : string;
 var t     : string;
       psn   : word;
       len   : byte;
       carry : boolean;
 begin
   Add[0] := #0;
   if Length(a) >= 254 then Exit;
   if Length(b) >= 254 then Exit;
   if a[0] = #0 then Exit;
   if b[0] = #0 then Exit;
   carry := FALSE;
   if Length(a) > Length(b) then 
     len := Succ(Length(a))
   else
     len := Succ(Length(b));
   a := LeftPad0(a, len);
   b := LeftPad0(b, len);
   FillChar(t[1], len, '0');
   t[0] := Char(len);
   psn  := Succ(len);
   { now add digits from left to right }
   while psn > 1 do
   begin
     Dec(psn);
     if carry then
       t[psn] := AddChar(Succ(a[psn]), b[psn], carry)
     else
       t[psn] := AddChar(a[psn], b[psn], carry);
   end;
   if carry then t[1] := '1';
   TrimLead0(t);
   if t = '' then t := '0';
   Add := t;
 end;

 function Compare(x, y : string) : shortint;
 { Returns -1 if x < y, 0 if equal, 1 if x > y }
 begin
   TrimLead0(x);  { remove any leading zeros }
   TrimLead0(y);
   if Length(x) = Length(y) then 
   begin
     if x = y then 
       Compare := 0
     else if x > y then
       Compare := 1
     else
       Compare := -1;
   end else if Length(x) > Length(y) then 
     Compare := 1
   else
     Compare := -1;
 end;

 function Sub(a, b : string) : string;
 var t             : string;
     psn, len      : word;
     borrow, minus : boolean;
 begin
   Sub[0] := #0;
   if Length(a) >= 254 then Exit;
   if Length(b) >= 254 then Exit;
   if a[0] = #0 then Exit;
   if b[0] = #0 then Exit;
   borrow := FALSE;
   minus  := FALSE;
   { Subtract smaller from larger }
   if Compare(a,b) = -1 then
   begin
     minus := TRUE;
     t := a;  a := b;  b := t;
   end;
   if Length(a) > Length(b) then 
     len := Succ(Length(a))
   else
     len := Succ(Length(b));
   a := LeftPad0(a, len);
   b := LeftPad0(b, len);
   FillChar(t[1], len, '0');
   t[0] := Char(len);
   psn  := Succ(len);
   { now subtract digits from right to left }
   while psn > 1 do
   begin
     Dec(psn);
     if borrow then
       t[psn] := SubChar(Pred(a[psn]), b[psn], borrow)
     else
       t[psn] := SubChar(a[psn], b[psn], borrow);
   end;
   TrimLead0(t);
   if t = '' then t := '0';
   if minus then
   begin
     Move(t[1], t[2], Length(t));
     t[1] := '-';
     Inc(t[0]);
   end;
   Sub := t;
 end;

 function Prod(a,b : string) : string;
 var t1, t2         : string;
     posn, times, n : word;
 begin
   prod[0] := #0;
   if Length(a) + Length(b) > 254 then Exit;
   if a[0] = #0 then Exit;
   if b[0] = #0 then Exit;
   { Multiply larger by smaller }
   if Compare(a,b) = -1 then
   begin
     t1 := a;  a := b;  b := t1;
   end;
   t2 := '0';
   { For each digit of multiplie, right to left, add together an
     appropriate number of copies of multiplicand, tack the right 
     number of zeros on the end, and add the result to the running
     total in t2 }
   for posn := Length(b) downto 1 do
   begin
     times := Ord(b[posn]) - 48;
     if times = 0 then t1 := '0'
     else begin
       t1 := a;
       for n := 2 to times do
         t1 := Add(t1, a);
     end;
     FillChar(t1[Succ(Length(t1))], Length(b) - posn, '0');
     Inc(t1[0], Length(b) - posn);
     t2 := Add(t2,t1);
   end;
   Prod := t2;
 end;

 function Divide(    a,b : string;
                 var rm  : string) : string;
 var t1, t2, t3 : string;
 begin
   Divide[0] := #0;
   rm[0] := #0;
   if a[0] = #0 then Exit;
   if b[0] = #0 then Exit;
   if Compare(a,b) = 0 then
   begin
     Divide := '1';
     rm     := '0';
   end else begin
     t1 := b;  t2 := '1';  t3 := '0';
     { While dividend is > t1, add zeros to t1 and to t2 }
     while Compare(a, t1) = 1 do
     begin
       Inc(t1[0]);  t1[Length(t1)] := '0';
       Inc(t2[0]);  t2[Length(t2)] := '0';
     end;
     { Get individual digits of quotient by repeated subtraction of t1.
       t1 is the divisor with a steadily decreasing number of zeroes after it }
     while Compare(t1, b) <> 0 do
     begin
       Dec(T1[0]);
       Dec(T2[0]);
       while Compare(a, t1) <> -1 do
       begin
         a := Sub(a, t1);
         if a[0] = #0 then Exit;
         t3 := Add(t3, t2);
         if t3[0] = #0 then Exit;
       end;
     end;
     Divide := t3;
     rm     := a;
   end;
 end;

 function Fact(var a : string) : string;
 var t1, t2 : string;
 begin
   t1 := '1';
   t2 := '1';
   if (a<> '1') AND (a <> '0') then
     while (t2 <> a) AND (t1[0] <> #0) do
     begin
       t2 := Add(t2, '1');
       t1 := Prod(t1, t2);
     end;
   Fact := t1;
 end;

 function Power(b,e : string) : string;
 var t1, t2, t3, rem : string;
 begin
   Power[0] := #0;
   if b[0] = #0 then Exit;
   if e[0] = #0 then Exit;
   power := '0';
   if b = '0' then Exit;
   power := '1';
   if e = '0' then Exit;
   t1 := b;
   t2 := e;
   t3 := '1';
   { Calculate power by halving and squaring }
   while (t2 <> '0') AND (t3[0] <> #0) do
   begin
     { Halve the exponent }
     t2 := Divide(t2, '2', rem);
     { If it was odd, multiply t3 by current value of t1 }
     if rem = '1' then t3 := Prod(t3, t1);
     { Square t1 }
     t1 := Prod(t1, t1);
   end;
   Power := t3;
 end;
end.
